#pragma once

#include <algorithm>
#include <limits>

#include <amdis/common/Math.hpp>
#include <amdis/operations/Arithmetic.hpp>
#include <amdis/operations/MaxMin.hpp>

#ifndef DOXYGEN

namespace Dune {

// some arithmetic operations with FieldVector

template <class T, int N>
FieldVector<T,N> operator-(FieldVector<T,N> v)
{
  return v *= -1;
}

template <class T, int N, class S,
  std::enable_if_t<std::is_arithmetic<S>::value,int> >
FieldVector<T,N> operator*(FieldVector<T,N> v, S factor)
{
  return v *= factor;
}

template <class S, class T, int N,
  std::enable_if_t<std::is_arithmetic<S>::value,int> >
FieldVector<T,N> operator*(S factor, FieldVector<T,N> v)
{
  return v *= factor;
}

template <class T, int N, class S,
  std::enable_if_t<std::is_arithmetic<S>::value,int> >
FieldVector<T,N> operator/(FieldVector<T,N> v, S factor)
{
  return v /= factor;
}

template <class T>
FieldVector<T,1> operator*(FieldVector<T,1> const& v, FieldVector<T,1> const& w)
{
  return {v[0] * w[0]};
}

template <class T, int N>
FieldVector<T,N> operator*(FieldVector<T,1> const& factor, FieldVector<T,N> v)
{
  return v *= factor[0];
}

template <class T, int N>
FieldVector<T,N> operator*(FieldVector<T,N> v, FieldVector<T,1> const& factor)
{
  return v *= factor[0];
}

// ----------------------------------------------------------------------------

/// Cross-product a 2d-vector = orthogonal vector
template <class T>
FieldVector<T, 2> cross(FieldVector<T, 2> const& a)
{
  return {{ a[1], -a[0] }};
}

/// Cross-product of two vectors (in 3d only)
template <class T>
FieldVector<T, 3> cross(FieldVector<T, 3> const& a, FieldVector<T, 3> const& b)
{
  return {{ a[1]*b[2] - a[2]*b[1],
            a[2]*b[0] - a[0]*b[2],
            a[0]*b[1] - a[1]*b[0] }};
}

/// Dot product (vec1^T * vec2)
template <class T, class S, int N>
auto dot(FieldVector<T,N> const& vec1, FieldVector<S,N> const& vec2)
{
  return vec1.dot(vec2);
}

template <class T, int N, int M,
  std::enable_if_t<( N!=1 && M!=1 ),int> >
auto operator*(FieldVector<T,N> const& v, FieldVector<T,M> const& w)
{
  static_assert(M == N, "Requires vectors of the same type!");
  return v.dot(w);
}

template <class T, class S, int N>
auto dot(FieldMatrix<T,1,N> const& vec1, FieldMatrix<S,1,N> const& vec2)
{
  return vec1[0].dot(vec2[0]);
}

// ----------------------------------------------------------------------------

namespace Impl
{
  template <class T, int N, class Operation>
  T accumulate(FieldVector<T, N> const& x, T init, Operation op)
  {
    for (int i = 0; i < N; ++i)
      init = op(init, x[i]);
    return init;
  }

  template <class T, int N, class Operation>
  T accumulate(FieldMatrix<T, 1, N> const& x, T init, Operation op)
  {
    for (int i = 0; i < N; ++i)
      init = op(init, x[0][i]);
    return init;
  }

} // end namespace Impl

/// Sum of vector entires.
template <class T, int N>
T sum(FieldVector<T, N> const& x)
{
  return Impl::accumulate(x, T(0), AMDiS::Operation::Plus{});
}

template <class T, int N>
T sum(FieldMatrix<T, 1, N> const& x)
{
  return Impl::accumulate(x, T(0), AMDiS::Operation::Plus{});
}


/// Dot-product with the vector itself
template <class T, int N>
auto unary_dot(FieldVector<T, N> const& x)
{
  auto op = [](auto const& a, auto const& b) { return a + AMDiS::Math::sqr(std::abs(b)); };
  return Impl::accumulate(x, T(0), op);
}

template <class T, int N>
auto unary_dot(FieldMatrix<T, 1, N> const& x)
{
  auto op = [](auto const& a, auto const& b) { return a + AMDiS::Math::sqr(std::abs(b)); };
  return Impl::accumulate(x, T(0), op);
}

/// Maximum over all vector entries
template <class T, int N>
auto max(FieldVector<T, N> const& x)
{
  return Impl::accumulate(x, std::numeric_limits<T>::lowest(), AMDiS::Operation::Max{});
}

template <class T, int N>
auto max(FieldMatrix<T, 1, N> const& x)
{
  return Impl::accumulate(x, std::numeric_limits<T>::lowest(), AMDiS::Operation::Max{});
}

/// Minimum over all vector entries
template <class T, int N>
auto min(FieldVector<T, N> const& x)
{
  return Impl::accumulate(x, std::numeric_limits<T>::max(), AMDiS::Operation::Min{});
}

template <class T, int N>
auto min(FieldMatrix<T, 1, N> const& x)
{
  return Impl::accumulate(x, std::numeric_limits<T>::max(), AMDiS::Operation::Min{});
}

/// Maximum of the absolute values of vector entries
template <class T, int N>
auto abs_max(FieldVector<T, N> const& x)
{
  return Impl::accumulate(x, T(0), AMDiS::Operation::AbsMax{});
}

template <class T, int N>
auto abs_max(FieldMatrix<T, 1, N> const& x)
{
  return Impl::accumulate(x, T(0), AMDiS::Operation::AbsMax{});
}

/// Minimum of the absolute values of vector entries
template <class T, int N>
auto abs_min(FieldVector<T, N> const& x)
{
  return Impl::accumulate(x, std::numeric_limits<T>::max(), AMDiS::Operation::AbsMin{});
}

template <class T, int N>
auto abs_min(FieldMatrix<T, 1, N> const& x)
{
  return Impl::accumulate(x, std::numeric_limits<T>::max(), AMDiS::Operation::AbsMin{});
}

// ----------------------------------------------------------------------------

/** \ingroup vector_norms
  *  \brief The 1-norm of a vector = sum_i |x_i|
  **/
template <class T, int N>
auto one_norm(FieldVector<T, N> const& x)
{
  auto op = [](auto const& a, auto const& b) { return a + std::abs(b); };
  return Impl::accumulate(x, T(0), op);
}

template <class T, int N>
auto one_norm(FieldMatrix<T, 1, N> const& x)
{
  auto op = [](auto const& a, auto const& b) { return a + std::abs(b); };
  return Impl::accumulate(x, T(0), op);
}

/** \ingroup vector_norms
  *  \brief The euklidean 2-norm of a vector = sqrt( sum_i |x_i|^2 )
  **/
template <class T, int N>
auto two_norm(FieldVector<T, N> const& x)
{
  return std::sqrt(unary_dot(x));
}

template <class T, int N>
auto two_norm(FieldMatrix<T, 1, N> const& x)
{
  return std::sqrt(unary_dot(x));
}

/** \ingroup vector_norms
  *  \brief The p-norm of a vector = ( sum_i |x_i|^p )^(1/p)
  **/
template <int p, class T, int N>
auto p_norm(FieldVector<T, N> const& x)
{
  auto op = [](auto const& a, auto const& b) { return a + AMDiS::Math::pow<p>(std::abs(b)); };
  return std::pow( Impl::accumulate(x, T(0), op), 1.0/p );
}

template <int p, class T, int N>
auto p_norm(FieldMatrix<T, 1, N> const& x)
{
  auto op = [](auto const& a, auto const& b) { return a + AMDiS::Math::pow<p>(std::abs(b)); };
  return std::pow( Impl::accumulate(x, T(0), op), 1.0/p );
}

/** \ingroup vector_norms
  *  \brief The infty-norm of a vector = max_i |x_i| = alias for \ref abs_max
  **/
template <class T, int N>
auto infty_norm(FieldVector<T, N> const& x)
{
  return abs_max(x);
}

template <class T, int N>
auto infty_norm(FieldMatrix<T, 1, N> const& x)
{
  return abs_max(x);
}

// ----------------------------------------------------------------------------

/// The euklidean distance between two vectors = |lhs-rhs|_2
template <class T, int N>
T distance(FieldVector<T, N> const& lhs, FieldVector<T, N> const& rhs)
{
  T result = 0;
  for (int i = 0; i < N; ++i)
    result += AMDiS::Math::sqr(lhs[i] - rhs[i]);
  return std::sqrt(result);
}

// ----------------------------------------------------------------------------

/// Outer product (vec1 * vec2^T)
template <class T, class S, int N, int M, int K>
auto outer(FieldMatrix<T,N,K> const& vec1, FieldMatrix<S,M,K> const& vec2)
{
  using result_type = FieldMatrix<decltype( std::declval<T>() * std::declval<S>() ), N, M>;
  result_type mat;
  for (int i = 0; i < N; ++i)
    for (int j = 0; j < M; ++j)
      mat[i][j] = vec1[i].dot(vec2[j]);
  return mat;
}

// ----------------------------------------------------------------------------

template <class T>
T det(FieldMatrix<T, 0, 0> const& /*mat*/)
{
  return 0;
}

/// Determinant of a 1x1 matrix
template <class T>
T det(FieldMatrix<T, 1, 1> const& mat)
{
  return mat[0][0];
}

/// Determinant of a 2x2 matrix
template <class T>
T det(FieldMatrix<T, 2, 2> const& mat)
{
  return mat[0][0]*mat[1][1] - mat[0][1]*mat[1][0];
}

/// Determinant of a 3x3 matrix
template <class T>
T det(FieldMatrix<T, 3, 3> const& mat)
{
  return mat[0][0]*mat[1][1]*mat[2][2] + mat[0][1]*mat[1][2]*mat[2][0] + mat[0][2]*mat[1][0]*mat[2][1]
      - (mat[0][2]*mat[1][1]*mat[2][0] + mat[0][1]*mat[1][0]*mat[2][2] + mat[0][0]*mat[1][2]*mat[2][1]);
}

/// Determinant of a NxN matrix
template <class T,  int N>
T det(FieldMatrix<T, N, N> const& mat)
{
  return mat.determinant();
}

/// Return the inverse of the matrix `mat`
template <class T, int N>
auto inv(FieldMatrix<T, N, N> mat)
{
  mat.invert();
  return mat;
}

/// Solve the linear system A*x = b
template <class T, int N>
void solve(FieldMatrix<T, N, N> const& A,  FieldVector<T, N>& x,  FieldVector<T, N> const& b)
{
  A.solve(x, b);
}


/// Gramian determinant = sqrt( det( DT^T * DF ) )
template <class T, int N, int M>
T gramian(FieldMatrix<T,N,M> const& DF)
{
  using std::sqrt;
  return sqrt( det(outer(DF, DF)) );
}

/// Gramian determinant, specialization for 1 column matrices
template <class T, int M>
T gramian(FieldMatrix<T, 1, M> const& DF)
{
  using std::sqrt;
  return sqrt(dot(DF[0], DF[0]));
}

// ----------------------------------------------------------------------------
// some arithmetic operations with FieldMatrix

template <class T, int M, int N>
FieldMatrix<T,N,M> trans(FieldMatrix<T, M, N> const& A)
{
  FieldMatrix<T,N,M> At;
  for (int i = 0; i < M; ++i)
    for (int j = 0; j < N; ++j)
      At[j][i] = A[i][j];

  return At;
}


template <class T, int M, int N, class S,
  std::enable_if_t<std::is_arithmetic<S>::value,int> >
FieldMatrix<T,M,N> operator*(S scalar, FieldMatrix<T, M, N> A)
{
  return A *= scalar;
}

template <class T, int M, int N, class S,
  std::enable_if_t<std::is_arithmetic<S>::value,int> >
FieldMatrix<T,M,N> operator*(FieldMatrix<T, M, N> A, S scalar)
{
  return A *= scalar;
}

template <class T, int N, int M>
FieldMatrix<T,N,M> operator*(FieldMatrix<T,1,1> scalar, FieldMatrix<T,N,M> mat)
{
  return mat *= scalar[0][0];
}

template <class T, int N, int M>
FieldMatrix<T,N,M> operator*(FieldMatrix<T,N,M> mat, FieldMatrix<T,1,1> scalar)
{
  return mat *= scalar[0][0];
}

template <class T, int N, int M>
FieldMatrix<T,N,M> operator*(FieldVector<T,1> scalar, FieldMatrix<T,N,M> mat)
{
  return mat *= scalar[0];
}

template <class T, int N>
FieldMatrix<T,N,1> operator*(FieldVector<T,1> scalar, FieldMatrix<T,N,1> mat)
{
  return mat *= scalar[0];
}

template <class T, int N, int M>
FieldMatrix<T,N,M> operator*(FieldMatrix<T,N,M> mat, FieldVector<T,1> scalar)
{
  return mat *= scalar[0];
}

template <class T, int N>
FieldMatrix<T,N,1> operator*(FieldMatrix<T,N,1> mat, FieldVector<T,1> scalar)
{
  return mat *= scalar[0];
}

template <class T, int M, int N, class S,
  std::enable_if_t<std::is_arithmetic<S>::value,int> >
FieldMatrix<T,M,N> operator/(FieldMatrix<T, M, N> A, S scalar)
{
  return A /= scalar;
}

template <class T, int M, int N>
FieldMatrix<T,M,N> operator+(FieldMatrix<T, M, N> A, FieldMatrix<T, M, N> const& B)
{
  return A += B;
}

template <class T, int M, int N>
FieldMatrix<T,M,N> operator-(FieldMatrix<T, M, N> A, FieldMatrix<T, M, N> const& B)
{
  return A -= B;
}


template <class T, int N, int M>
FieldVector<T,N> operator*(FieldMatrix<T,N,M> const& mat, FieldVector<T,M> const& vec)
{
  return Dune::FMatrixHelp::mult(mat, vec);
}



template <class T, int M, int N, int L>
FieldMatrix<T,M,N> multiplies(FieldMatrix<T, M, L> const& A,  FieldMatrix<T, L, N> const& B)
{
  return A.rightmultiplyany(B);
}



template <class T, int M, int N, int L>
FieldMatrix<T,M,N> multiplies_AtB(FieldMatrix<T, L, M> const& A,  FieldMatrix<T, N, L> const& B)
{
  FieldMatrix<T,M,N> C;

  for (int m = 0; m < M; ++m) {
    for (int n = 0; n < N; ++n) {
      C[m][n] = 0;
      for (int l = 0; l < L; ++l)
        C[m][n] += A[l][m] * B[n][l];
    }
  }
  return C;
}

template <class T, int M, int N, int L>
FieldMatrix<T,M,N> multiplies_ABt(FieldMatrix<T, M, L> const& A,  FieldMatrix<T, N, L> const& B)
{
  FieldMatrix<T,M,N> C;

  for (int m = 0; m < M; ++m) {
    for (int n = 0; n < N; ++n) {
      C[m][n] = 0;
      for (int l = 0; l < L; ++l)
        C[m][n] += A[m][l] * B[n][l];
    }
  }
  return C;
}

template <class T, int M, int N, int L>
FieldMatrix<T,M,N>& multiplies_ABt(FieldMatrix<T, M, L> const& A,  FieldMatrix<T, N, L> const& B, FieldMatrix<T,M,N>& C)
{
  for (int m = 0; m < M; ++m) {
    for (int n = 0; n < N; ++n) {
      C[m][n] = 0;
      for (int l = 0; l < L; ++l)
        C[m][n] += A[m][l] * B[n][l];
    }
  }
  return C;
}

template <class T, int M, int N>
FieldMatrix<T,M,N>& multiplies_ABt(FieldMatrix<T, M, N> const& A,  DiagonalMatrix<T, N> const& B, FieldMatrix<T,M,N>& C)
{
  for (int m = 0; m < M; ++m) {
    for (int n = 0; n < N; ++n) {
      C[m][n] = A[m][n] * B.diagonal(n);
    }
  }
  return C;
}


template <class T, int N>
T const& at(FieldMatrix<T,N,1> const& vec, std::size_t i)
{
  return vec[i][0];
}

template <class T, int M>
T const& at(FieldMatrix<T,1,M> const& vec, std::size_t i)
{
  return vec[0][i];
}

template <class T, int N>
T const& at(FieldVector<T,N> const& vec, std::size_t i)
{
  return vec[i];
}

} // end namespace AMDiS

#endif
