// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:

#include <iostream>
#include <type_traits>

#include <amdis/AMDiS.hpp>
#include <amdis/ProblemStat.hpp>
#include <amdis/gridfunctions/DiscreteFunction.hpp>
#include <amdis/gridfunctions/DOFVectorView.hpp>

#include "Tests.hpp"

using namespace AMDiS;

using ElliptParam   = YaspGridBasis<2, 2>;
using ElliptProblem = ProblemStat<ElliptParam>;

template <class GB, class T>
bool comp(DOFVector<GB,T> const& U, DOFVector<GB,T> const& V)
{
  if (U.size() != V.size())
    return false;

  using Int = typename DOFVector<GB,T>::size_type;
  for (Int i = 0; i < U.size(); ++i) {
    if (U[i] != V[i])
      return false;
  }
  return true;
}

int main(int argc, char** argv)
{
  AMDiS::init(argc, argv);

  using namespace Dune::Indices;

  ElliptProblem prob("ellipt");
  prob.initialize(INIT_ALL);

  // create 3 copies of the solution vector
  auto U0 = prob.solutionVector();
  auto U1 = prob.solutionVector();
  auto U2 = prob.solutionVector();

  auto u0 = makeDOFVectorView(U0);
  auto u1 = makeDOFVectorView(U1);
  auto u2 = makeDOFVectorView(U2);

  auto expr = [](auto const& x) { return x[0] + x[1]; };
  u0.interpolate_noalias(expr);
  u1.interpolate(expr);
  u2 << expr;

  AMDIS_TEST( comp(U0, U1) );
  AMDIS_TEST( comp(U0, U2) );

  auto expr2 = [](auto const& x) { return 2*x[0] - 3*x[1]; };

  u0.interpolate_noalias(u2 + expr2);
  u1.interpolate(u1 + expr2);
  u2 += expr2;

  AMDIS_TEST( comp(U0, U1) );
  AMDIS_TEST( comp(U0, U2) );

  auto expr3 = [](auto const& x) { return -0.5*x[0] - 2*x[1]; };

  u0.interpolate_noalias(u2 - expr3);
  u1.interpolate(u1 - expr3);
  u2 -= expr3;

  AMDIS_TEST( comp(U0, U1) );
  AMDIS_TEST( comp(U0, U2) );

  auto du0 = derivative(u0);
  auto localFct = localFunction(u0);
  auto dlocalFct = derivative(localFct);
  for (auto const& e : elements(prob.gridView()))
  {
    localFct.bind(e);
    auto geo = e.geometry();
    auto local = referenceElement(geo).position(0,0);
    auto y0 = localFct(local);
    auto y1 = u0(geo.global(local));

    AMDIS_TEST(std::abs(y0 - y1) < 1.e-10);
    localFct.unbind();

    dlocalFct.bind(e);
    auto g0 = dlocalFct(local);
    auto g1 = du0(geo.global(local));

    AMDIS_TEST(two_norm(g0 - g1) < 1.e-10);
    dlocalFct.unbind();
  }

  auto V0 = makeDOFVector<float>(prob.globalBasis());
  auto v0 = makeDOFVectorView(V0);
  v0 << expr;

  AMDiS::finalize();
  return 0;
}
